/*=============================================================================
    Copyright (c) 2016 Kohei Takahashi

    Distributed under the Boost Software License, Version 1.0. (See accompanying
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#ifndef boost_phoenix_core_detail_index_sequence_hpp
#define boost_phoenix_core_detail_index_sequence_hpp

#include <boost/fusion/support/detail/index_sequence.hpp>

namespace boost { namespace phoenix { namespace detail {
using ::boost::fusion::detail::index_sequence;
using ::boost::fusion::detail::make_index_sequence;
}}}

#endif
