#ifndef STAN_LANG_AST_NODE_OMNI_IDX_HPP
#define STAN_LANG_AST_NODE_OMNI_IDX_HPP

namespace stan {
  namespace lang {

    /**
     * AST structure for representing all legal indexes.
     */
    struct omni_idx {
      /**
       * Construct an omni-index.
       */
      omni_idx();
    };

  }
}
#endif
