// Copyright David Abrahams 2002.
// Distributed under the Boost Software License, Version 1.0. (See
// accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
#ifndef OVERLOADS_DWA2002101_HPP
# define OVERLOADS_DWA2002101_HPP

# include <boost/python/detail/prefix.hpp>

# include <boost/python/detail/overloads_fwd.hpp>
# include <boost/python/detail/defaults_def.hpp>

#endif // OVERLOADS_DWA2002101_HPP
